@@ -34,6 +34,80 @@ MessageDefinitionPrinter::MessageDefinitionPrinter(const Descriptor *message,
 
 MessageDefinitionPrinter::~MessageDefinitionPrinter() = default;
 
+void MessageDefinitionPrinter::printClassMembers()
+{
+    Indent();
+    common::iterateMessageFields(
+            m_descriptor, [&](const FieldDescriptor *field, const PropertyMap &propertyMap) {
+                if (common::isOneofField(field))
+                    return;
+                if (common::isOptionalField(field)) {
+                    m_printer->Print(propertyMap, CommonTemplates::MemberOptionalTemplate());
+                } else if (common::isPureMessage(field)) {
+                    m_printer->Print(propertyMap, CommonTemplates::MemberMessageTemplate());
+                } else if (field->is_repeated() && !field->is_map()) {
+                    m_printer->Print(propertyMap, CommonTemplates::MemberRepeatedTemplate());
+                } else {
+                    m_printer->Print(propertyMap, CommonTemplates::MemberTemplate());
+                }
+            });
+
+    common::iterateOneofFields(
+            m_descriptor, [&](const OneofDescriptor *, const PropertyMap &propertyMap) {
+                m_printer->Print(propertyMap, CommonTemplates::MemberOneofTemplate());
+            });
+    Outdent();
+}
+
+void MessageDefinitionPrinter::printDataClass()
+{
+    m_printer->Print(m_typeMap, CommonTemplates::ClassMessageDataBeginDeclarationTemplate());
+    printPublicBlock();
+    printDataClassConstructor();
+    printDataClassCopy();
+    printClassMembers();
+    encloseClass();
+    m_printer->Print("\n");
+}
+
+void MessageDefinitionPrinter::printDataClassConstructor()
+{
+    Indent();
+    m_printer->Print(m_typeMap, CommonTemplates::ConstructorMessageDataDefinitionTemplate());
+    printInitializationList();
+    m_printer->Print(CommonTemplates::EmptyBracesTemplate());
+    Outdent();
+}
+
+void MessageDefinitionPrinter::printDataClassCopy()
+{
+    Indent();
+    m_printer->Print(m_typeMap, CommonTemplates::CopyConstructorMessageDataDefinitionTemplate());
+    Indent();
+    m_printer->Indent();
+    common::iterateMessageFields(
+            m_descriptor, [&](const FieldDescriptor *field, const PropertyMap &propertyMap) {
+                if (common::isOneofField(field) || common::isOptionalField(field))
+                    return;
+
+                m_printer->Print(",\n");
+                m_printer->Print(propertyMap,
+                                 common::isPureMessage(field)
+                                         ? CommonTemplates::CopyInitializerMemberMessageTemplate()
+                                         : CommonTemplates::CopyInitializerMemberTemplate());
+            });
+    common::iterateOneofFields(
+            m_descriptor, [&](const OneofDescriptor *, const PropertyMap &propertyMap) {
+                m_printer->Print(",\n");
+                m_printer->Print(propertyMap,
+                                 CommonTemplates::CopyInitializerMemberOneofTemplate());
+            });
+    m_printer->Outdent();
+    Outdent();
+    m_printer->Print(CommonTemplates::EmptyBracesTemplate());
+    Outdent();
+}
+
 void MessageDefinitionPrinter::printClassDefinitionPrivate()
 {
     if (common::hasNestedMessages(m_descriptor)) {
@@ -46,6 +120,8 @@ void MessageDefinitionPrinter::printClassDefinitionPrivate()
         m_printer->Print(scopeNamespaces, CommonTemplates::NamespaceClosingTemplate());
     }
 
+    printDataClass();
+
     printDestructor();
     printFieldsOrdering();
     printRegisterBody();
@@ -54,6 +130,7 @@ void MessageDefinitionPrinter::printClassDefinitionPrivate()
     printMoveSemantic();
     printComparisonOperators();
     printGetters();
+    printPublicExtras();
 }
 
 void MessageDefinitionPrinter::printClassDefinition()
@@ -67,10 +144,10 @@ void MessageDefinitionPrinter::printRegisterBody()
 
     m_printer->Print(m_typeMap, CommonTemplates::MetaTypeRegistrationMessageDefinition());
     Indent();
-    if (Options::instance().hasQml()) {
-        m_printer->Print(m_typeMap, CommonTemplates::MetaTypeRegistrationQmlListTemplate());
-        m_printer->Print(m_typeMap, CommonTemplates::QmlRegisterTypeTemplate());
-    }
+
+    // Special case:
+    if (m_descriptor->full_name() == "google.protobuf.Any")
+        m_printer->Print("QT_PREPEND_NAMESPACE(QtProtobuf)::Any::registerTypes();\n");
 
     common::iterateMessageFields(
             m_descriptor, [&](const FieldDescriptor *field, const PropertyMap &propertyMap) {
@@ -164,15 +241,21 @@ void MessageDefinitionPrinter::printUintData(const char *templateString)
     // JSon data starts where string metadata ends
     size_t jsonOffset = metaCharDataSize() + NullTerminator;
     const int numFields = m_descriptor->field_count();
-    for (int i = 0; i < numFields; ++i) {
+    for (int i = 0, propertyIndex = 0; i < numFields; ++i, ++propertyIndex) {
         const FieldDescriptor *field = m_descriptor->field(i);
         const std::map<std::string, std::string> variables = {
             { "json_name_offset", std::to_string(jsonOffset) },
             { "field_number", std::to_string(field->number()) },
-            { "property_index", std::to_string(i) },
+            { "property_index", std::to_string(propertyIndex) },
             { "field_flags", common::collectFieldFlags(field) },
             { "json_name", field->json_name() },
         };
+
+        // Oneof and optional properties generate additional has<FieldName> property next to the
+        // field property one.
+        if (common::isOneofField(field) || common::isOptionalField(field))
+            ++propertyIndex;
+
         m_printer->Print(variables, templateString);
 
         const size_t length = field->json_name().length();
@@ -218,69 +301,25 @@ size_t MessageDefinitionPrinter::charDataSize() const
 void MessageDefinitionPrinter::printConstructors()
 {
     m_printer->Print(m_typeMap, CommonTemplates::ConstructorMessageDefinitionTemplate());
-    printInitializationList();
     m_printer->Print(CommonTemplates::EmptyBracesTemplate());
-
-    if (m_descriptor->full_name() == "google.protobuf.Timestamp") {
-        m_printer->Print(
-                "Timestamp::Timestamp(const QDateTime &datetime) : "
-                "QProtobufMessage(&Timestamp::staticMetaObject),"
-                "m_seconds(datetime.toMSecsSinceEpoch() / 1000)\n"
-                ", m_nanos((datetime.toMSecsSinceEpoch() % 1000) * 1000)\n"
-                "{}\n"
-                "Timestamp::operator QDateTime() const\n"
-                "{\n"
-                "    return QDateTime::fromMSecsSinceEpoch(m_seconds * 1000 + m_nanos / 1000);\n"
-                "}\n");
-    }
 }
 
 void MessageDefinitionPrinter::printInitializationList()
 {
-    const auto numFields = m_descriptor->field_count();
     Indent();
-    for (int i = 0; i < numFields; ++i) {
-        const FieldDescriptor *field = m_descriptor->field(i);
-        auto propertyMap = common::producePropertyMap(field, m_descriptor);
-        if (!field->is_repeated() && !field->is_map()) {
-            switch (field->type()) {
-            case FieldDescriptor::TYPE_DOUBLE:
-            case FieldDescriptor::TYPE_FLOAT:
-                propertyMap["initializer"] = "0.0";
-                break;
-            case FieldDescriptor::TYPE_FIXED32:
-            case FieldDescriptor::TYPE_FIXED64:
-            case FieldDescriptor::TYPE_INT32:
-            case FieldDescriptor::TYPE_INT64:
-            case FieldDescriptor::TYPE_SINT32:
-            case FieldDescriptor::TYPE_SINT64:
-            case FieldDescriptor::TYPE_UINT32:
-            case FieldDescriptor::TYPE_UINT64:
-                propertyMap["initializer"] = "0";
-                break;
-            case FieldDescriptor::TYPE_BOOL:
-                propertyMap["initializer"] = "false";
-                break;
-            case FieldDescriptor::TYPE_ENUM:
-                propertyMap["initializer"] = propertyMap["scope_type"]
-                        + "::" + field->enum_type()->value(0)->name();
-                break;
-            default:
-                propertyMap["initializer"] = "";
-                break;
-            }
-        }
+    m_printer->Indent();
+    common::iterateMessageFields(
+            m_descriptor, [&](const FieldDescriptor *field, PropertyMap propertyMap) {
+                if (field->is_repeated() || common::isOneofField(field)
+                    || common::isOptionalField(field))
+                    return;
 
-        if (common::isPureMessage(field)) {
-            m_printer->Print(",\n");
-            m_printer->Print(propertyMap, CommonTemplates::InitializerMemberMessageTemplate());
-        } else {
-            if (!propertyMap["initializer"].empty()) {
-                m_printer->Print(",\n");
-                m_printer->Print(propertyMap, CommonTemplates::InitializerMemberTemplate());
-            }
-        }
-    }
+                if (!propertyMap["initializer"].empty()) {
+                    m_printer->Print(",\n");
+                    m_printer->Print(propertyMap, CommonTemplates::InitializerMemberTemplate());
+                }
+            });
+    m_printer->Outdent();
     Outdent();
 }
 
@@ -288,98 +327,14 @@ void MessageDefinitionPrinter::printCopyFunctionality()
 {
     assert(m_descriptor != nullptr);
     m_printer->Print(m_typeMap, CommonTemplates::CopyConstructorDefinitionTemplate());
-    Indent();
-    common::iterateMessageFields(
-            m_descriptor, [&](const FieldDescriptor *field, const PropertyMap &propertyMap) {
-                m_printer->Print(",\n");
-                m_printer->Print(propertyMap,
-                                 common::isPureMessage(field)
-                                         ? CommonTemplates::InitializerMemberMessageTemplate()
-                                         : CommonTemplates::CopyInitializerMemberTemplate());
-            });
-    Outdent();
-
-    m_printer->Print("\n{\n");
-
-    Indent();
-    common::iterateMessageFields(
-            m_descriptor, [&](const FieldDescriptor *field, const PropertyMap &propertyMap) {
-                if (common::isPureMessage(field)) {
-                    m_printer->Print(propertyMap, CommonTemplates::CopyMemberMessageTemplate());
-                }
-            });
-    Outdent();
-    m_printer->Print(CommonTemplates::SimpleBlockEnclosureTemplate());
-
     m_printer->Print(m_typeMap, CommonTemplates::AssignmentOperatorDefinitionTemplate());
-    Indent();
-    common::iterateMessageFields(
-                m_descriptor, [&](const FieldDescriptor *field, const PropertyMap &propertyMap) {
-        m_printer->Print(propertyMap,
-                         common::isPureMessage(field) ?
-                             CommonTemplates::AssignMemberMessageTemplate() :
-                             CommonTemplates::CopyMemberTemplate());
-    });
-    m_printer->Print(CommonTemplates::AssignmentOperatorReturnTemplate());
-    Outdent();
-    m_printer->Print(CommonTemplates::SimpleBlockEnclosureTemplate());
 }
 
 void MessageDefinitionPrinter::printMoveSemantic()
 {
     assert(m_descriptor != nullptr);
     m_printer->Print(m_typeMap, CommonTemplates::MoveConstructorDefinitionTemplate());
-    common::iterateMessageFields(
-                m_descriptor, [&](const FieldDescriptor *field, const PropertyMap &propertyMap) {
-        if (common::isPureMessage(field)) {
-            m_printer->Print(",\n");
-            m_printer->Print(propertyMap,
-                             CommonTemplates::InitializerMemberMessageTemplate());
-        }
-    });
-    m_printer->Print("\n{\n");
-
-    Indent();
-    common::iterateMessageFields(
-                m_descriptor, [&](const FieldDescriptor *field, const PropertyMap &propertyMap) {
-        if (field->type() == FieldDescriptor::TYPE_MESSAGE
-                || field->type() == FieldDescriptor::TYPE_STRING
-                || field->type() == FieldDescriptor::TYPE_BYTES || field->is_repeated()) {
-            m_printer->Print(propertyMap,
-                             common::isPureMessage(field) ?
-                                 CommonTemplates::MoveMemberMessageTemplate() :
-                                 CommonTemplates::MoveConstructorMemberComplexTemplate());
-        } else {
-            m_printer->Print(propertyMap,
-                             (field->type() != FieldDescriptor::TYPE_ENUM) ?
-                                 CommonTemplates::MoveMemberTemplate() :
-                                 CommonTemplates::MoveMemberEnumTemplate());
-        }
-    });
-    Outdent();
-    m_printer->Print(CommonTemplates::SimpleBlockEnclosureTemplate());
-
     m_printer->Print(m_typeMap, CommonTemplates::MoveAssignmentOperatorDefinitionTemplate());
-    Indent();
-    common::iterateMessageFields(
-                m_descriptor, [&](const FieldDescriptor *field, const PropertyMap &propertyMap) {
-        if (field->type() == FieldDescriptor::TYPE_MESSAGE
-                || field->type() == FieldDescriptor::TYPE_STRING
-                || field->type() == FieldDescriptor::TYPE_BYTES || field->is_repeated()) {
-            m_printer->Print(propertyMap,
-                             common::isPureMessage(field) ?
-                                 CommonTemplates::MoveAssignMemberMessageTemplate() :
-                                 CommonTemplates::MoveAssignMemberComplexTemplate());
-        } else {
-            m_printer->Print(propertyMap,
-                             (field->type() != FieldDescriptor::TYPE_ENUM) ?
-                                 CommonTemplates::MoveMemberTemplate() :
-                                 CommonTemplates::MoveMemberEnumTemplate());
-        }
-    });
-    m_printer->Print(CommonTemplates::AssignmentOperatorReturnTemplate());
-    Outdent();
-    m_printer->Print(CommonTemplates::SimpleBlockEnclosureTemplate());
 }
 
 void MessageDefinitionPrinter::printComparisonOperators()
@@ -392,6 +347,8 @@ void MessageDefinitionPrinter::printComparisonOperators()
     Indent();
     common::iterateMessageFields(
                 m_descriptor, [&](const FieldDescriptor *field, PropertyMap &propertyMap) {
+        if (field->containing_oneof())
+            return;
         m_printer->Print("\n&& ");
         if (common::isPureMessage(field)) {
             m_printer->Print(propertyMap,
@@ -405,6 +362,13 @@ void MessageDefinitionPrinter::printComparisonOperators()
                              CommonTemplates::EqualOperatorMemberTemplate());
         }
     });
+
+    common::iterateOneofFields(
+            m_descriptor, [&](const OneofDescriptor *, const PropertyMap &propertyMap) {
+                m_printer->Print("\n&& ");
+                m_printer->Print(propertyMap, CommonTemplates::EqualOperatorMemberOneofTemplate());
+            });
+
     Outdent();
     Outdent();
 
@@ -417,45 +381,102 @@ void MessageDefinitionPrinter::printComparisonOperators()
 void MessageDefinitionPrinter::printGetters()
 {
     common::iterateMessageFields(
-                m_descriptor, [&](const FieldDescriptor *field, PropertyMap &propertyMap) {
-        if (common::isPureMessage(field)) {
-            m_printer->Print(propertyMap,
-                             CommonTemplates::PrivateGetterMessageDefinitionTemplate());
-            m_printer->Print(propertyMap,
-                             CommonTemplates::GetterMessageDefinitionTemplate());
-        }
-        if (field->is_repeated()) {
-            if (field->type() == FieldDescriptor::TYPE_MESSAGE && !field->is_map()
-                    && !common::isQtType(field) && Options::instance().hasQml()) {
-                m_printer->Print(propertyMap,
-                                 CommonTemplates::GetterQmlListDefinitionTemplate());
-            }
-        }
-    });
+            m_descriptor, [&](const FieldDescriptor *field, PropertyMap &propertyMap) {
+                if (common::isOneofField(field)) {
+                    if (common::isPureMessage(field)) {
+                        m_printer->Print(
+                                propertyMap,
+                                CommonTemplates::PrivateGetterOneofMessageDefinitionTemplate());
+                        m_printer->Print(propertyMap,
+                                         CommonTemplates::GetterOneofMessageDefinitionTemplate());
+                    } else {
+                        m_printer->Print(propertyMap,
+                                         CommonTemplates::PrivateGetterOneofDefinitionTemplate());
+                        m_printer->Print(propertyMap,
+                                         CommonTemplates::GetterOneofDefinitionTemplate());
+                    }
+                    return;
+                }
+
+                if (common::isOptionalField(field)) {
+                    m_printer->Print(propertyMap,
+                                     CommonTemplates::PrivateGetterOptionalDefinitionTemplate());
+                    m_printer->Print(propertyMap,
+                                     CommonTemplates::GetterOptionalDefinitionTemplate());
+                    return;
+                }
+
+                if (common::hasQmlAlias(field)) {
+                    m_printer->Print(propertyMap, CommonTemplates::GetterNonScriptableDefinitionTemplate());
+                }
+
+                if (common::isPureMessage(field)) {
+                    m_printer->Print(propertyMap,
+                                     CommonTemplates::PrivateGetterMessageDefinitionTemplate());
+                    m_printer->Print(propertyMap,
+                                     CommonTemplates::GetterMessageDefinitionTemplate());
+                    m_printer->Print(propertyMap,
+                                     CommonTemplates::ClearMessageDefinitionTemplate());
+                } else {
+                    m_printer->Print(propertyMap, CommonTemplates::GetterDefinitionTemplate());
+                }
+                if (field->is_repeated()) {
+                    m_printer->Print(propertyMap,
+                                     CommonTemplates::GetterComplexDefinitionTemplate());
+                }
+            });
 
     common::iterateMessageFields(
-                m_descriptor, [&](const FieldDescriptor *field, PropertyMap &propertyMap) {
-        switch (field->type()) {
-        case FieldDescriptor::TYPE_MESSAGE:
-            if (!field->is_map() && !field->is_repeated() && !common::isQtType(field)) {
+            m_descriptor, [&](const FieldDescriptor *field, PropertyMap &propertyMap) {
+                if (common::isOneofField(field)) {
+                    m_printer->Print(propertyMap, CommonTemplates::SetterOneofDefinitionTemplate());
+                    m_printer->Print(
+                            propertyMap,
+                            common::isPureMessage(field)
+                                    ? CommonTemplates::PrivateSetterOneofMessageDefinitionTemplate()
+                                    : CommonTemplates::PrivateSetterOneofDefinitionTemplate());
+                    return;
+                }
+                if (common::isOptionalField(field)) {
+                    m_printer->Print(propertyMap,
+                                     CommonTemplates::SetterOptionalDefinitionTemplate());
+                    m_printer->Print(propertyMap,
+                                     CommonTemplates::PrivateSetterOptionalDefinitionTemplate());
+                    m_printer->Print(propertyMap,
+                                     CommonTemplates::ClearOptionalDefinitionTemplate());
+                    return;
+                }
+                if (common::hasQmlAlias(field)) {
+                    m_printer->Print(propertyMap, CommonTemplates::SetterNonScriptableDefinitionTemplate());
+                }
+                switch (field->type()) {
+                case FieldDescriptor::TYPE_MESSAGE:
+                    if (common::isPureMessage(field)) {
+                        m_printer->Print(propertyMap,
+                                         CommonTemplates::PrivateSetterMessageDefinitionTemplate());
+                        m_printer->Print(propertyMap,
+                                         CommonTemplates::SetterMessageDefinitionTemplate());
+                    } else {
+                        m_printer->Print(propertyMap,
+                                         CommonTemplates::SetterComplexDefinitionTemplate());
+                    }
+                    break;
+                case FieldDescriptor::FieldDescriptor::TYPE_STRING:
+                case FieldDescriptor::FieldDescriptor::TYPE_BYTES:
+                    m_printer->Print(propertyMap,
+                                     CommonTemplates::SetterComplexDefinitionTemplate());
+                    break;
+                default:
+                    m_printer->Print(propertyMap, CommonTemplates::SetterDefinitionTemplate());
+                    break;
+                }
+            });
+    common::iterateOneofFields(
+            m_descriptor, [&](const OneofDescriptor *, const PropertyMap &propertyMap) {
                 m_printer->Print(propertyMap,
-                                 CommonTemplates::PrivateSetterMessageDefinitionTemplate());
-                m_printer->Print(propertyMap,
-                                 CommonTemplates::SetterMessageDefinitionTemplate());
-            } else {
-                m_printer->Print(propertyMap,
-                                 CommonTemplates::SetterComplexDefinitionTemplate());
-            }
-            break;
-        case FieldDescriptor::FieldDescriptor::TYPE_STRING:
-        case FieldDescriptor::FieldDescriptor::TYPE_BYTES:
-            m_printer->Print(propertyMap,
-                             CommonTemplates::SetterComplexDefinitionTemplate());
-            break;
-        default:
-            break;
-        }
-    });
+                                 CommonTemplates::GetterOneofFieldNumberDefinitionTemplate());
+                m_printer->Print(propertyMap, CommonTemplates::ClearOneofDefinitionTemplate());
+            });
 }
 
 void MessageDefinitionPrinter::printDestructor()
@@ -463,6 +484,25 @@ void MessageDefinitionPrinter::printDestructor()
     m_printer->Print(m_typeMap, "$classname$::~$classname$() = default;\n\n");
 }
 
+void MessageDefinitionPrinter::printPublicExtras()
+{
+    if (m_descriptor->full_name() == "google.protobuf.Timestamp") {
+        m_printer->Print(
+                "Timestamp Timestamp::fromDateTime(const QDateTime &dateTime)\n"
+                "{\n"
+                "    Timestamp ts;\n"
+                "    ts.setSeconds(dateTime.toMSecsSinceEpoch() / 1000);\n"
+                "    ts.setNanos((dateTime.toMSecsSinceEpoch() % 1000) * 1000000);\n"
+                "    return ts;\n"
+                "}\n\n"
+                "QDateTime Timestamp::toDateTime() const\n"
+                "{\n"
+                "    return QDateTime::fromMSecsSinceEpoch(\n"
+                "            seconds() * 1000 + nanos() / 1000000, QTimeZone(QTimeZone::UTC));\n"
+                "}\n\n");
+    }
+}
+
 void MessageDefinitionPrinter::printClassRegistration(Printer *printer)
 {
     if (common::hasNestedMessages(m_descriptor)) {
@@ -478,3 +518,10 @@ void MessageDefinitionPrinter::printClassRegistration(Printer *printer)
 
     printer->Print(m_typeMap, CommonTemplates::RegistrarTemplate());
 }
+
+void MessageDefinitionPrinter::printQmlPluginClassRegistration()
+{
+    Indent();
+    m_printer->Print(m_typeMap, CommonTemplates::QmlRegisterMessageTypeTemplate());
+    Outdent();
+}
