@@ -12,10 +12,16 @@
 
 #include <qtprotobuftestscommon.h>
 
+using namespace Qt::Literals::StringLiterals;
+
 class QtProtobufTypesGenerationTest : public QObject
 {
     Q_OBJECT
 private slots:
+    void initTestCase()
+    {
+        qRegisterProtobufTypes();
+    }
     void EmptyMessageTest();
     void BoolMessageTest();
     void IntMessageTest();
@@ -33,6 +39,9 @@ private slots:
     void DoubleMessageTest();
     void ComplexMessageTest();
     void BytesMessageTest();
+    void OneofMessageEmptyTest();
+    void OneofMessageTest();
+    void OneofMessageCopyComplexValueTest();
 
     void AssignmentOperatorTest();
     void MoveOperatorTest();
@@ -65,14 +74,14 @@ void QtProtobufTypesGenerationTest::BoolMessageTest()
     QCOMPARE(test.property(propertyName).value<bool>(), true);
     QCOMPARE(test.testFieldBool(), true);
 
-    QCOMPARE(SimpleBoolMessage::TestFieldBoolProtoFieldNumber, 1);
+    QCOMPARE(int(SimpleBoolMessage::QtProtobufFieldEnum::TestFieldBoolProtoFieldNumber), 1);
     QCOMPARE(test.propertyOrdering.getMessageFullName(),
              "qtprotobufnamespace.tests.SimpleBoolMessage");
 }
 
 void QtProtobufTypesGenerationTest::IntMessageTest()
 {
-    const char *propertyName = "testFieldInt" QT_PROTOBUF_PROPERTY_SUFFIX;
+    const char *propertyName = "testFieldInt";
     qProtobufAssertMessagePropertyRegistered<qtprotobufnamespace::tests::SimpleIntMessage, QtProtobuf::int32>(1, "QtProtobuf::int32", propertyName);
 
     qtprotobufnamespace::tests::SimpleIntMessage test;
@@ -80,7 +89,9 @@ void QtProtobufTypesGenerationTest::IntMessageTest()
     QCOMPARE(test.property(propertyName).value<QtProtobuf::int32>(), 1);
     QCOMPARE(test.testFieldInt(), 1);
 
-    QCOMPARE(qtprotobufnamespace::tests::SimpleIntMessage::TestFieldIntProtoFieldNumber, 1);
+    QCOMPARE(int(qtprotobufnamespace::tests::SimpleIntMessage::QtProtobufFieldEnum::
+                         TestFieldIntProtoFieldNumber),
+             1);
     QCOMPARE(test.propertyOrdering.getMessageFullName(),
              "qtprotobufnamespace.tests.SimpleIntMessage");
 }
@@ -95,7 +106,7 @@ void QtProtobufTypesGenerationTest::SIntMessageTest()
     QCOMPARE(test.property(propertyName).value<QtProtobuf::sint32>(), 1);
     QCOMPARE(test.testFieldInt(), 1);
 
-    QCOMPARE(SimpleSIntMessage::TestFieldIntProtoFieldNumber, 1);
+    QCOMPARE(int(SimpleSIntMessage::QtProtobufFieldEnum::TestFieldIntProtoFieldNumber), 1);
     QCOMPARE(test.propertyOrdering.getMessageFullName(),
              "qtprotobufnamespace.tests.SimpleSIntMessage");
 }
@@ -110,7 +121,7 @@ void QtProtobufTypesGenerationTest::UIntMessageTest()
     QCOMPARE(test.property(propertyName).value<QtProtobuf::uint32>(), (uint32_t)1);
     QCOMPARE(test.testFieldInt(), 1u);
 
-    QCOMPARE(SimpleUIntMessage::TestFieldIntProtoFieldNumber, 1);
+    QCOMPARE(int(SimpleUIntMessage::QtProtobufFieldEnum::TestFieldIntProtoFieldNumber), 1);
     QCOMPARE(test.propertyOrdering.getMessageFullName(),
              "qtprotobufnamespace.tests.SimpleUIntMessage");
 }
@@ -125,7 +136,7 @@ void QtProtobufTypesGenerationTest::Int64MessageTest()
     QCOMPARE(test.property(propertyName).value<QtProtobuf::int64>(), 1);
     QCOMPARE(test.testFieldInt(), 1);
 
-    QCOMPARE(SimpleInt64Message::TestFieldIntProtoFieldNumber, 1);
+    QCOMPARE(int(SimpleInt64Message::QtProtobufFieldEnum::TestFieldIntProtoFieldNumber), 1);
     QCOMPARE(test.propertyOrdering.getMessageFullName(),
              "qtprotobufnamespace.tests.SimpleInt64Message");
 }
@@ -140,7 +151,7 @@ void QtProtobufTypesGenerationTest::SInt64MessageTest()
     QCOMPARE(test.property(propertyName).value<QtProtobuf::sint64>(), 1);
     QCOMPARE(test.testFieldInt(), 1);
 
-    QCOMPARE(SimpleSInt64Message::TestFieldIntProtoFieldNumber, 1);
+    QCOMPARE(int(SimpleSInt64Message::QtProtobufFieldEnum::TestFieldIntProtoFieldNumber), 1);
     QCOMPARE(test.propertyOrdering.getMessageFullName(),
              "qtprotobufnamespace.tests.SimpleSInt64Message");
 }
@@ -155,14 +166,14 @@ void QtProtobufTypesGenerationTest::UInt64MessageTest()
     QCOMPARE(test.property(propertyName).value<QtProtobuf::uint64>(), 1u);
     QCOMPARE(test.testFieldInt(), 1u);
 
-    QCOMPARE(SimpleUInt64Message::TestFieldIntProtoFieldNumber, 1u);
+    QCOMPARE(uint(SimpleUInt64Message::QtProtobufFieldEnum::TestFieldIntProtoFieldNumber), 1u);
     QCOMPARE(test.propertyOrdering.getMessageFullName(),
              "qtprotobufnamespace.tests.SimpleUInt64Message");
 }
 
 void QtProtobufTypesGenerationTest::FixedInt32MessageTest()
 {
-    const char *propertyName = "testFieldFixedInt32" QT_PROTOBUF_PROPERTY_SUFFIX;
+    const char *propertyName = "testFieldFixedInt32";
     qProtobufAssertMessagePropertyRegistered<SimpleFixedInt32Message, QtProtobuf::fixed32>(1, "QtProtobuf::fixed32", propertyName);
 
     SimpleFixedInt32Message test;
@@ -170,7 +181,9 @@ void QtProtobufTypesGenerationTest::FixedInt32MessageTest()
     QCOMPARE(test.property(propertyName).value<QtProtobuf::fixed32>(), 1u);
     QCOMPARE(test.testFieldFixedInt32(), 1u);
 
-    QCOMPARE(SimpleFixedInt32Message::TestFieldFixedInt32ProtoFieldNumber, 1u);
+    QCOMPARE(
+            uint(SimpleFixedInt32Message::QtProtobufFieldEnum::TestFieldFixedInt32ProtoFieldNumber),
+            1u);
     QCOMPARE(test.propertyOrdering.getMessageFullName(),
              "qtprotobufnamespace.tests.SimpleFixedInt32Message");
 }
@@ -185,14 +198,16 @@ void QtProtobufTypesGenerationTest::FixedInt64MessageTest()
     QCOMPARE(test.property(propertyName).value<QtProtobuf::fixed64>(), 1u);
     QCOMPARE(test.testFieldFixedInt64(), 1u);
 
-    QCOMPARE(SimpleFixedInt64Message::TestFieldFixedInt64ProtoFieldNumber, 1u);
+    QCOMPARE(
+            uint(SimpleFixedInt64Message::QtProtobufFieldEnum::TestFieldFixedInt64ProtoFieldNumber),
+            1u);
     QCOMPARE(test.propertyOrdering.getMessageFullName(),
              "qtprotobufnamespace.tests.SimpleFixedInt64Message");
 }
 
 void QtProtobufTypesGenerationTest::SFixedInt32MessageTest()
 {
-    const char *propertyName = "testFieldFixedInt32" QT_PROTOBUF_PROPERTY_SUFFIX;
+    const char *propertyName = "testFieldFixedInt32";
     qProtobufAssertMessagePropertyRegistered<SimpleSFixedInt32Message, QtProtobuf::sfixed32>(1, "QtProtobuf::sfixed32", propertyName);
 
     SimpleSFixedInt32Message test;
@@ -200,7 +215,9 @@ void QtProtobufTypesGenerationTest::SFixedInt32MessageTest()
     QCOMPARE(test.property(propertyName).value<QtProtobuf::sfixed32>(), 1);
     QCOMPARE(test.testFieldFixedInt32(), 1);
 
-    QCOMPARE(SimpleSFixedInt32Message::TestFieldFixedInt32ProtoFieldNumber, 1);
+    QCOMPARE(
+            int(SimpleSFixedInt32Message::QtProtobufFieldEnum::TestFieldFixedInt32ProtoFieldNumber),
+            1);
     QCOMPARE(test.propertyOrdering.getMessageFullName(),
              "qtprotobufnamespace.tests.SimpleSFixedInt32Message");
 }
@@ -215,7 +232,9 @@ void QtProtobufTypesGenerationTest::SFixedInt64MessageTest()
     QCOMPARE(test.property(propertyName).value<QtProtobuf::sfixed64>(), 1);
     QCOMPARE(test.testFieldFixedInt64(), 1);
 
-    QCOMPARE(SimpleSFixedInt64Message::TestFieldFixedInt64ProtoFieldNumber, 1);
+    QCOMPARE(
+            int(SimpleSFixedInt64Message::QtProtobufFieldEnum::TestFieldFixedInt64ProtoFieldNumber),
+            1);
     QCOMPARE(test.propertyOrdering.getMessageFullName(),
              "qtprotobufnamespace.tests.SimpleSFixedInt64Message");
 }
@@ -229,10 +248,10 @@ void QtProtobufTypesGenerationTest::StringMessageTest()
     QCOMPARE(SimpleStringMessage::staticMetaObject.property(propertyNumber).typeId(), QMetaType::QString);
     QCOMPARE(SimpleStringMessage::staticMetaObject.property(propertyNumber).name(), propertyName);
     QVERIFY(test.setProperty(propertyName, QVariant::fromValue(QString("test1"))));
-    QCOMPARE(test.property(propertyName).toString(), QStringLiteral("test1"));
-    QCOMPARE(test.testFieldString(), QStringLiteral("test1"));
+    QCOMPARE(test.property(propertyName).toString(), "test1"_L1);
+    QCOMPARE(test.testFieldString(), "test1"_L1);
 
-    QCOMPARE(SimpleStringMessage::TestFieldStringProtoFieldNumber, 6);
+    QCOMPARE(int(SimpleStringMessage::QtProtobufFieldEnum::TestFieldStringProtoFieldNumber), 6);
     QCOMPARE(test.propertyOrdering.getMessageFullName(),
              "qtprotobufnamespace.tests.SimpleStringMessage");
 }
@@ -244,14 +263,15 @@ void QtProtobufTypesGenerationTest::FloatMessageTest()
     int index = SimpleFloatMessage::propertyOrdering.indexOfFieldNumber(7);
     int propertyNumber = SimpleFloatMessage::propertyOrdering.getPropertyIndex(index); //See simpletest.proto
     QCOMPARE(SimpleFloatMessage::staticMetaObject.property(propertyNumber).typeId(), QMetaType::Float);
-    QCOMPARE(QLatin1String(SimpleFloatMessage::staticMetaObject.property(propertyNumber).name()), QStringLiteral("testFieldFloat"));
+    QCOMPARE(SimpleFloatMessage::staticMetaObject.property(propertyNumber).name(),
+             "testFieldFloat"_L1);
 
     float assignedValue = 1.55f;
     QVERIFY(test.setProperty(propertyName, QVariant::fromValue<float>(assignedValue)));
     QCOMPARE(test.property(propertyName).toFloat(), assignedValue);
     QCOMPARE(test.testFieldFloat(), assignedValue);
 
-    QCOMPARE(SimpleFloatMessage::TestFieldFloatProtoFieldNumber, 7);
+    QCOMPARE(int(SimpleFloatMessage::QtProtobufFieldEnum::TestFieldFloatProtoFieldNumber), 7);
     QCOMPARE(test.propertyOrdering.getMessageFullName(),
              "qtprotobufnamespace.tests.SimpleFloatMessage");
 }
@@ -270,14 +290,14 @@ void QtProtobufTypesGenerationTest::DoubleMessageTest()
     QCOMPARE(test.property(propertyName).toDouble(), assignedValue);
     QCOMPARE(test.testFieldDouble(), assignedValue);
 
-    QCOMPARE(SimpleDoubleMessage::TestFieldDoubleProtoFieldNumber, 8);
+    QCOMPARE(int(SimpleDoubleMessage::QtProtobufFieldEnum::TestFieldDoubleProtoFieldNumber), 8);
     QCOMPARE(test.propertyOrdering.getMessageFullName(),
              "qtprotobufnamespace.tests.SimpleDoubleMessage");
 }
 
 void QtProtobufTypesGenerationTest::ComplexMessageTest()
 {
-    const char *propertyName = "testComplexField";
+    const char *propertyName = "testComplexField_p";
     qProtobufAssertMessagePropertyRegistered<qtprotobufnamespace::tests::ComplexMessage, qtprotobufnamespace::tests::SimpleStringMessage*>(
                 2, "qtprotobufnamespace::tests::SimpleStringMessage*", propertyName);
 
@@ -318,6 +338,59 @@ void QtProtobufTypesGenerationTest::BytesMessageTest()
              "qtprotobufnamespace.tests.SimpleBytesMessage");
 }
 
+void QtProtobufTypesGenerationTest::OneofMessageEmptyTest()
+{
+    OneofMessage test;
+    QVERIFY(!test.hasTestOneofFieldInt());
+    QVERIFY(!test.hasTestOneofComplexField());
+    QCOMPARE(test.testOneofField(), OneofMessage::TestOneofFields::UninitializedField);
+}
+
+void QtProtobufTypesGenerationTest::OneofMessageTest()
+{
+    OneofMessage test;
+
+    test.setTestOneofFieldInt(10);
+    QVERIFY(test.hasTestOneofFieldInt());
+    QVERIFY(!test.hasTestOneofComplexField());
+    QVERIFY(!test.hasTestOneofSecondComplexField());
+    QCOMPARE(test.testOneofFieldInt(), 10);
+    QCOMPARE(test.testOneofField(), OneofMessage::TestOneofFields::TestOneofFieldInt);
+
+    ComplexMessage complexField;
+    SimpleStringMessage stringField;
+    stringField.setTestFieldString("Qt Test");
+    complexField.setTestFieldInt(20);
+    complexField.setTestComplexField(stringField);
+    test.setTestOneofComplexField(complexField);
+    QVERIFY(!test.hasTestOneofFieldInt());
+    QVERIFY(test.hasTestOneofComplexField());
+    QVERIFY(!test.hasTestOneofSecondComplexField());
+    QCOMPARE(test.testOneofComplexField(), complexField);
+    QCOMPARE(test.testOneofField(), OneofMessage::TestOneofFields::TestOneofComplexField);
+}
+
+void QtProtobufTypesGenerationTest::OneofMessageCopyComplexValueTest()
+{
+    OneofMessage test;
+    ComplexMessage complexField;
+    SimpleStringMessage stringField;
+    stringField.setTestFieldString("Qt Test");
+    complexField.setTestFieldInt(20);
+    complexField.setTestComplexField(stringField);
+    test.setTestOneofComplexField(complexField);
+    QVERIFY(test.hasTestOneofComplexField());
+    QVERIFY(!test.hasTestOneofSecondComplexField());
+    QCOMPARE(test.testOneofComplexField(), complexField);
+    QCOMPARE(test.testOneofField(), OneofMessage::TestOneofFields::TestOneofComplexField);
+
+    test.setTestOneofSecondComplexField(test.testOneofComplexField());
+    QVERIFY(!test.hasTestOneofComplexField());
+    QVERIFY(test.hasTestOneofSecondComplexField());
+    QCOMPARE(test.testOneofSecondComplexField(), complexField);
+    QCOMPARE(test.testOneofField(), OneofMessage::TestOneofFields::TestOneofSecondComplexField);
+}
+
 void QtProtobufTypesGenerationTest::AssignmentOperatorTest()
 {
     const char *propertyName = "testFieldInt";
@@ -341,16 +414,15 @@ void QtProtobufTypesGenerationTest::MoveOperatorTest()
     const char *propertyName = "testFieldInt";
     qtprotobufnamespace::tests::SimpleIntMessage test;
     qtprotobufnamespace::tests::SimpleIntMessage test2;
-    test2.setTestFieldInt(35);
-
-    qtprotobufnamespace::tests::SimpleIntMessage test3(std::move(test2));
-    test2.setTestFieldInt(35);
-
+    test2.setTestFieldInt(25);
     test.setProperty(propertyName, QVariant::fromValue<QtProtobuf::int32>(15));
-    test.setTestFieldInt(25);
     test = std::move(test2);
-    QCOMPARE(test.testFieldInt(), 35);
-    QCOMPARE(test2.testFieldInt(), 0);
+    QCOMPARE(test.testFieldInt(), 25);
+
+    qtprotobufnamespace::tests::SimpleIntMessage test3;
+    test3.setTestFieldInt(35);
+    qtprotobufnamespace::tests::SimpleIntMessage test4(std::move(test3));
+    QCOMPARE(test4.testFieldInt(), 35);
 }
 
 void QtProtobufTypesGenerationTest::AccessMessageFieldsFromGetter()
